import os
import shutil
from functools import wraps

import pytest

import autofit as af
from autofit import Paths
from autofit.optimize.non_linear.output import AbstractOutput
from test_autofit.mock import MockClassNLOx4, MockClassNLOx5, MockClassNLOx6

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class MockOutput(AbstractOutput):
    def __init__(
        self,
        model,
        paths,
        most_probable_vector=None,
        most_likely_vector=None,
        vector_at_sigma=None,
        sample_vector=None,
    ):

        super(MockOutput, self).__init__(model=model, paths=paths)

        self._most_probable_vector = most_probable_vector
        self._most_likely_vector = most_likely_vector
        self._vector_at_sigma = vector_at_sigma
        self._sample_vector = sample_vector

    @property
    def most_probable_vector(self):
        return self._most_probable_vector

    @property
    def most_likely_vector(self):
        return self._most_likely_vector

    def vector_at_sigma(self, sigma):
        return [(sigma * value[0], sigma * value[1]) for value in self._vector_at_sigma]

    def vector_from_sample_index(self, sample_index):
        return self._sample_vector[sample_index]


class TestOutput:
    def test__most_probable_instance(self):

        model = af.ModelMapper(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_vector=[
                1.0,
                2.0,
                3.0,
                4.0,
                -5.0,
                -6.0,
                -7.0,
                -8.0,
                9.0,
                10.0,
            ],
        )

        most_probable = output.most_probable_instance

        assert most_probable.mock_class_1.one == 1.0
        assert most_probable.mock_class_1.two == 2.0
        assert most_probable.mock_class_1.three == 3.0
        assert most_probable.mock_class_1.four == 4.0

        assert most_probable.mock_class_2.one == (-5.0, -6.0)
        assert most_probable.mock_class_2.two == (-7.0, -8.0)
        assert most_probable.mock_class_2.three == 9.0
        assert most_probable.mock_class_2.four == 10.0

    def test__most_likely_instance(self):

        model = af.ModelMapper(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

        output = MockOutput(
            model=model,
            paths=Paths(),
            most_likely_vector=[
                21.0,
                22.0,
                23.0,
                24.0,
                25.0,
                -26.0,
                -27.0,
                28.0,
                29.0,
                30.0,
            ],
        )

        most_likely = output.most_likely_instance

        assert most_likely.mock_class_1.one == 21.0
        assert most_likely.mock_class_1.two == 22.0
        assert most_likely.mock_class_1.three == 23.0
        assert most_likely.mock_class_1.four == 24.0

        assert most_likely.mock_class_2.one == (25.0, -26.0)
        assert most_likely.mock_class_2.two == (-27.0, 28.0)
        assert most_likely.mock_class_2.three == 29.0
        assert most_likely.mock_class_2.four == 30.0

    def test__vector_at_upper_and_lower_sigma(self,):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_vector=[1.0, 2.0, 3.0, 4.1],
            vector_at_sigma=[(0.88, 1.12), (1.88, 2.12), (2.88, 3.12), (3.88, 4.12)],
        )

        params_upper = output.vector_at_upper_sigma(sigma=1.0)
        assert params_upper == pytest.approx([1.12, 2.12, 3.12, 4.12], 1e-2)
        params_lower = output.vector_at_lower_sigma(sigma=1.0)
        assert params_lower == pytest.approx([0.88, 1.88, 2.88, 3.88], 1e-2)

        params_upper = output.vector_at_upper_sigma(sigma=2.0)
        assert params_upper == pytest.approx(
            [2.0 * 1.12, 2.0 * 2.12, 2.0 * 3.12, 2.0 * 4.12], 1e-2
        )
        params_lower = output.vector_at_lower_sigma(sigma=2.0)
        assert params_lower == pytest.approx(
            [2.0 * 0.88, 2.0 * 1.88, 2.0 * 2.88, 2.0 * 3.88], 1e-2
        )

        instance = output.instance_at_upper_sigma(sigma=1.0)
        assert instance.mock_class.one == pytest.approx(1.12, 1e-2)
        assert instance.mock_class.two == pytest.approx(2.12, 1e-2)
        assert instance.mock_class.three == pytest.approx(3.12, 1e-2)
        assert instance.mock_class.four == pytest.approx(4.12, 1e-2)

        instance = output.instance_at_lower_sigma(sigma=1.0)
        assert instance.mock_class.one == pytest.approx(0.88, 1e-2)
        assert instance.mock_class.two == pytest.approx(1.88, 1e-2)
        assert instance.mock_class.three == pytest.approx(2.88, 1e-2)
        assert instance.mock_class.four == pytest.approx(3.88, 1e-2)

    def test__gaussian_priors(self):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_vector=[1.0, 2.0, 3.0, 4.1],
            vector_at_sigma=[(0.88, 1.12), (1.88, 2.12), (2.88, 3.12), (3.88, 4.12)],
        )

        gaussian_priors = output.gaussian_priors_at_sigma(sigma=1.0)

        assert gaussian_priors[0][0] == 1.0
        assert gaussian_priors[1][0] == 2.0
        assert gaussian_priors[2][0] == 3.0
        assert gaussian_priors[3][0] == 4.1

        assert gaussian_priors[0][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[1][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[2][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[3][1] == pytest.approx(0.22, 1e-2)

    def test__offset_vector_from_input_vector(self):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model, paths=Paths(), most_probable_vector=[1.0, -2.0, 3.0, 4.0]
        )

        offset_values = output.offset_vector_from_input_vector(
            input_vector=[1.0, 1.0, 2.0, 3.0]
        )

        assert offset_values == [0.0, -3.0, 1.0, 1.0]

    def test__instance_from_sample_index(self,):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            sample_vector=[[1.0, -2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]],
        )

        instance = output.instance_from_sample_index(sample_index=0)

        assert instance.mock_class.one == 1.0
        assert instance.mock_class.two == -2.0
        assert instance.mock_class.three == 3.0
        assert instance.mock_class.four == 4.0

        instance = output.instance_from_sample_index(sample_index=1)

        assert instance.mock_class.one == 5.0
        assert instance.mock_class.two == 6.0
        assert instance.mock_class.three == 7.0
        assert instance.mock_class.four == 8.0

    def test__error_vector_at_sigma(self):
        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            vector_at_sigma=[(0.88, 1.12), (1.88, 2.12), (2.88, 3.12), (3.88, 4.12)],
        )

        errors = output.error_vector_at_sigma(sigma=1.0)
        assert errors == pytest.approx(
            [1.12 - 0.88, 2.12 - 1.88, 3.12 - 2.88, 4.12 - 3.88], 1e-2
        )

        errors_instance = output.error_instance_at_sigma(sigma=1.0)
        assert errors_instance.mock_class.one == pytest.approx(1.12 - 0.88, 1e-2)
        assert errors_instance.mock_class.two == pytest.approx(2.12 - 1.88, 1e-2)
        assert errors_instance.mock_class.three == pytest.approx(3.12 - 2.88, 1e-2)
        assert errors_instance.mock_class.four == pytest.approx(4.12 - 3.88, 1e-2)

        errors = output.error_vector_at_sigma(sigma=2.0)
        assert errors == pytest.approx(
            [
                2.0 * (1.12 - 0.88),
                2.0 * (2.12 - 1.88),
                2.0 * (3.12 - 2.88),
                2.0 * (4.12 - 3.88),
            ],
            1e-2,
        )

    def test__errors_at_upper_and_lower_sigma(self):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_vector=[1.1, 2.0, 3.0, 4.0],
            vector_at_sigma=[(0.88, 1.12), (1.88, 2.12), (2.88, 3.12), (3.88, 4.12)],
        )

        upper_errors = output.error_vector_at_upper_sigma(sigma=1.0)
        assert upper_errors == pytest.approx([0.02, 0.12, 0.12, 0.12], 1e-2)

        errors_instance = output.error_instance_at_upper_sigma(sigma=1.0)
        assert errors_instance.mock_class.one == pytest.approx(0.02, 1e-2)
        assert errors_instance.mock_class.two == pytest.approx(0.12, 1e-2)
        assert errors_instance.mock_class.three == pytest.approx(0.12, 1e-2)
        assert errors_instance.mock_class.four == pytest.approx(0.12, 1e-2)

        upper_errors = output.error_vector_at_upper_sigma(sigma=2.0)
        assert upper_errors == pytest.approx([1.14, 2.24, 3.24, 4.24], 1e-2)

        lower_errors = output.error_vector_at_lower_sigma(sigma=1.0)
        assert lower_errors == pytest.approx([0.22, 0.12, 0.12, 0.12], 1e-2)

        errors_instance = output.error_instance_at_lower_sigma(sigma=1.0)
        assert errors_instance.mock_class.one == pytest.approx(0.22, 1e-2)
        assert errors_instance.mock_class.two == pytest.approx(0.12, 1e-2)
        assert errors_instance.mock_class.three == pytest.approx(0.12, 1e-2)
        assert errors_instance.mock_class.four == pytest.approx(0.12, 1e-2)

        lower_errors = output.error_vector_at_lower_sigma(sigma=0.5)
        assert lower_errors == pytest.approx([0.66, 1.06, 1.56, 2.06], 1e-2)



import pytest

import autofit as af


class Class:
    def __init__(self, argument):
        self.argument = argument


def fitness_function(instance: Class):
    return -(instance.argument - 30) ** 2


@pytest.mark.parametrize(
    "optimizer",
    [
        af.MultiNest
    ]
)
def test_simple(optimizer):
    model = af.Model(
        Class,
        argument=af.UniformPrior(10, 100)
    )
    result = optimizer.simple_fit(
        model,
        fitness_function
    )

    assert result.instance.argument == pytest.approx(30, abs=1.0)
    assert result.likelihood == pytest.approx(0, abs=1)



from abc import ABC, abstractmethod
from typing import List, Iterator

from .phase_output import PhaseOutput


class AttributePredicate:
    def __init__(self, attribute: str):
        """
        Used to produce predicate objects for filtering in the aggregator.

        When an unrecognised attribute is called on an aggregator an instance
        of this object is created. This object implements comparison methods
        facilitating construction of predicates.

        Parameters
        ----------
        attribute
            The name of the attribute this predicate relates to.
        """
        self.attribute = attribute

    def __eq__(self, value):
        """
        Create a predicate which asks whether the given value is equal to
        the attribute of a phase.
        """
        return EqualityPredicate(
            self.attribute,
            value
        )

    def __ne__(self, other):
        """
        Create a predicate which asks whether the given value is not equal to
        the attribute of a phase.
        """
        return ~(self == other)

    def contains(self, value):
        """
        Create a predicate which asks whether the given is contained within
        the attribute of a phase.
        """
        return ContainsPredicate(
            self.attribute,
            value
        )


class AbstractPredicate(ABC):
    """
    Comparison between a value and some attribute of a phase
    """

    def filter(
            self,
            phases: List[PhaseOutput]
    ) -> Iterator[PhaseOutput]:
        """
        Only return phases for which this predicate evaluates to True

        Parameters
        ----------
        phases

        Returns
        -------

        """
        return filter(
            lambda phase: self(phase),
            phases
        )

    def __invert__(self) -> "NotPredicate":
        """
        A predicate that evaluates to True when this predicate evaluates
        to False
        """
        return NotPredicate(
            self
        )

    @abstractmethod
    def __call__(self, phase: PhaseOutput) -> bool:
        """
        Does the attribute of the phase match the requirement of this predicate?
        """


class ComparisonPredicate(AbstractPredicate, ABC):
    def __init__(
            self,
            attribute: str,
            value
    ):
        """
        Compare an attribute of a phase with a value.

        Parameters
        ----------
        attribute
            An attribute of a phase
        value
            A value to which the attribute is compared
        """
        self.attribute = attribute
        self.value = value


class ContainsPredicate(ComparisonPredicate):
    def __call__(
            self,
            phase: PhaseOutput
    ) -> bool:
        """
        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True iff the value of the attribute of the phase contains
        the value associated with this predicate
        """
        return self.value in getattr(
            phase,
            self.attribute
        )


class EqualityPredicate(ComparisonPredicate):
    def __call__(self, phase):
        """
        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True iff the value of the attribute of the phase is equal to
        the value associated with this predicate
        """
        return getattr(
            phase,
            self.attribute
        ) == self.value


class NotPredicate(AbstractPredicate):
    def __init__(
            self,
            predicate: AbstractPredicate
    ):
        """
        Negates the output of a predicate.

        If the predicate would have returned True for a given phase
        it now returns False and vice-versa.

        Parameters
        ----------
        predicate
            A predicate that is negated
        """
        self.predicate = predicate

    def __call__(self, phase: PhaseOutput) -> bool:
        """
        Evaluate the predicate for the phase and return the negation
        of the result.

        Parameters
        ----------
        phase
            The output of an AutoFit phase

        Returns
        -------
        The negation of the underlying predicate
        """
        return not self.predicate(
            phase
        )
